/*
 * $Id: arch-linux-distrib.h,v 1.1.2.3 1997/11/17 20:31:38 randall Exp $
 */
/*
 * $Log: arch-linux-distrib.h,v $
 * Revision 1.1.2.3  1997/11/17 20:31:38  randall
 * Got linux port up to date.
 *
 * Revision 1.1.2.2  1997/10/14 20:25:37  randall
 * First working version of distributed cilk on DEC machines.
 *
 * Revision 1.1.2.1  1997/06/16 19:48:48  randall
 * Added linux to supported architectures.
 *
 * Revision 1.2  1997/03/01 17:42:30  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#define CACHE_LINE 64

/* padding, to be used in structures to avoid false sharing */
#define CILK_CACHE_LINE_PAD  char __dummy[CACHE_LINE]

#define CACHE_LINE_ALIGNED(decl) decl  __attribute__((aligned(CACHE_LINE)))

#if (!CILK_LINUX_DISTRIB)
#error CILK_LINUX_DISTRIB not defined !
#endif

/***********************************************************\
 * Lock implementation
\***********************************************************/

/* a lock variable */
typedef volatile int Cilk_mutex;

/***********************************************************\
 * Machine size
\***********************************************************/

#include <unistd.h>

static inline int Cilk_partition_size(void)
{
     /* 
      * This is only really right for an unloaded machine.
      */
     return 1;
}

/*
 * Get OS page size.
 */
static inline size_t Cilk_get_page_size(void)
{
     return sysconf(_SC_PAGESIZE);
}

/***********************************************************
 * Interaction with unix
 ***********************************************************/
#define Cilk_yield()     { usleep(10000); }

/***********************************************************\
 * Various types of membars
\***********************************************************/


/*
 *  Ensure that all previous memory operations are completed before
 *  continuing.
 */
static inline void Cilk_fence(void)
{
     /* unknown --- I just have a one-processor machine */
}

/*
 *  Ensure that all previous writes are globally visible before any
 *  future writes become visible.
 */
static inline void Cilk_membar_StoreStore(void)
{
     /* unknown --- I just have a one-processor machine */
}

/*
 *  Ensure that all previous writes are globally visible before any
 *  future reads are performed.
 */
static inline void Cilk_membar_StoreLoad(void)
{
     /* unknown --- I just have a one-processor machine */
}

/***********************************************************\
 * Timing functions
\***********************************************************/

#include <sys/times.h>

/* type to hold a high-resolution timer.  Must be an arithmetic type. */
typedef long Cilk_time;

/* set after program starts */
DECLARE_PRIVATE_VARIABLE(extern Cilk_time, wall_start, );

static inline void Cilk_init_timers(void)
{
     struct tms buf;
     USE_PRIVATE(wall_start) = times(&buf);
}

static inline Cilk_time Cilk_get_time(void)
{
     struct tms buf;
     times(&buf);
     
     return (buf.tms_utime + buf.tms_stime);
}

static inline double Cilk_time_to_sec(Cilk_time t)
{
     return (double) t / 100.0;
}

static inline Cilk_time Cilk_get_wall_time(void)
{
     struct tms buf;
     
     return times(&buf) - USE_PRIVATE(wall_start);
}

static inline double Cilk_wall_time_to_sec(Cilk_time t)
{
     return (double) t / 100.0;
}

/***********************************************************\
 * Other random includes
\***********************************************************/

#include <signal.h>

/***********************************************************\
 * Typedefs for messaging library
\***********************************************************/

#define MAXMACHINES 100
#define MAX_SEND_QUEUE 1024
#define MAX_MISSING 128

/* header for packets in transit */
typedef struct
{
  void (*handler)();         /* active message function      */
  int size;                  /* # of bytes in message body   */
  int first_seq_no;          /* first seq no of this message */
  int sender;                /* message sender               */
  int seq_no;                /* sequence number from sender  */
  int ack_seq_no;            /* piggy-backed acknowledgement */
} TransitHeader;

/* header for messages waiting to be sent */
typedef struct SendMessage
{
  struct SendMessage *next;
  void (*handler)();         /* active message function      */
  int size;                  /* # of bytes in message body   */
  int buf_size;              /* # of bytes to free when done */
  int seq_no_of_first_chunk; /* sequence # of first chunk    */
  int dest;                  /* destination machine          */
  void *data;                /* data of message (body)       */
} SendMessage;

/* header for packets being reconstructed at receiver */
typedef struct RecvMessage
{
  struct RecvMessage *next;
  int size;                  /* # of bytes in message body   */
  int seq_no_of_first_chunk; /* sequence # of first chunk    */
  int chunks_so_far;         /* # of chunks received         */
  void *data;                /* data of message (body)       */
} RecvMessage;

/* data type for keeping track of which packets need to be sent */
typedef struct
{
  int seq_no;
  SendMessage *msg;
} SendReq;

/* data type for keeping track of which packets we still need from sender */
typedef struct
{
  int sender;                /* message sender               */
  int seq_no;                /* sequence # of missing chunk  */
  Cilk_time gone_missing;    /* time we noticed absence      */
} MissingPacket;

/***********************************************************\
 * Typedefs for cache
\***********************************************************/

/*
 * This number should be the size of the largest virtually indexed
 * cache in the system, divided by its associativity and the
 * system page size.  For example, on SuperSparc processors, there
 * is a 16k virtually indexed direct-mapped L1 cache, and a page
 * size of 8k.  This gives a color value of (16k / 1 / 8k) = 2.
 */
/* Is this right for Linux machines?
 */
#define VCOLORS 2
