/*
 * $Id: arch-solaris-distrib.c,v 1.1.2.10 1997/12/19 18:55:52 randall Exp $
 */
/*
 * $Log: arch-solaris-distrib.c,v $
 * Revision 1.1.2.10  1997/12/19 18:55:52  randall
 * Got v8 version working.
 *
 * Revision 1.1.2.9  1997/12/18 23:14:14  randall
 * Cache eviction fix, some Makefile changes.
 *
 * Revision 1.1.2.8  1997/11/13 20:56:07  randall
 * More debug fixup.
 *
 * Revision 1.1.2.7  1997/11/13 20:43:39  randall
 * Debugging support enhanced.
 *
 * Revision 1.1.2.6  1997/11/13 18:51:12  randall
 * Update of solaris UDP version.
 *
 * Revision 1.1.2.5  1997/10/21 23:02:34  randall
 * Separated out UDP implementation.  Reorganized caching algorithm.
 *
 * Revision 1.1.2.4  1997/10/14 20:47:02  randall
 * Fixed to make sure distributed version still works on xolas.
 *
 * Revision 1.1.2.3  1997/06/18 21:37:59  randall
 * Made host determination more robust.
 *
 * Revision 1.1.2.2  1997/06/18 16:10:59  randall
 * Fixed bug: if HOST is not defined, problems happen
 *
 * Revision 1.1.2.1  1997/06/13 22:33:01  randall
 * Added distributed version to the repository under
 * branch "distributed"
 *
 * Revision 1.9  1996/12/13 14:06:01  athena
 * Added support for Solaris threads
 *
 * Revision 1.8  1996/12/12 23:27:53  athena
 * Cilk2c now understands `register int foo asm("r0")' and the like.
 * Cilk2c understands #pragma segments [+-].
 * Added flexibility to the rts to accomodate the above.
 *
 * Revision 1.7  1996/12/11 20:06:03  athena
 * Simplified management of shared/private segments.  Now Milk
 * manages both segments in the same way, and for both there
 * is the option of using the linker or of manually handling
 * them.
 *
 * Revision 1.6  1996/12/11 19:19:56  athena
 * The RTS now manages its own segments.  Fixed some incorrect
 * initialization sequences.
 *
 * Revision 1.5  1996/12/11 17:17:41  athena
 * *** empty log message ***
 *
 * Revision 1.4  1996/12/10 21:41:52  athena
 * Massive renaming of variables to ease port to architechtures
 * whose address space is not so flexible as on Solaris.
 * All global variables are now either SHARED, PRIVATE or SHARED-ReadOnly.
 * All accesses are wrapped with macros, and (almost) all variables
 * are initialized.  The net effect is that it should be easy to
 * put all shared variables in a structure, and things like that.
 *
 * Revision 1.3  1996/12/04 19:10:41  athena
 * Another little optimization of lock release
 * (suggested by the manual)
 *
 * Revision 1.2  1996/12/04 18:55:21  athena
 * Changed memory locks to use CAS as opposed to SWAP (SWAP
 * is deprecated in V9)
 *
 * Revision 1.1.1.1  1996/12/03 22:20:57  athena
 * Created repository for Cilk 5 (a.k.a. Milk)
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <stdarg.h>
#include <stdio.h>
#include <cilk.h>
#include <cilk-internal.h>
#include <cilk-lib.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/priocntl.h>
#include <sys/tspriocntl.h>
#include <sys/procfs.h>
#include <signal.h>
#include <stdio.h>
#include <netdb.h>
#include <unistd.h>
#include <assert.h>
#include <string.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <stropts.h>
#include <poll.h>
#include <fcntl.h>

/***********************************************************\
 * Debugging & tracing
\***********************************************************/

DECLARE_MACHINE_VARIABLE(static Cilk_mutex, dprintf_lock, );

static void wait_dprintf_lock(void)
{
     long long start = gethrtime();
     while(!Cilk_mutex_try(&USE_MACHINE(dprintf_lock)))
       {
	 if (gethrtime() - start > 100000000)  /* 100ms timeout */
	   break;
       }
}

void Cilk_dprintf(const char *fmt,...)
{
     va_list l;

     /* ensure nice interleaving of debugging messages */
     wait_dprintf_lock();
     fprintf(stderr, "   (%2d,%2d): ", USE_MACHINE(Machine), USE_PRIVATE(Process));
     va_start(l, fmt);
     vfprintf(stderr, fmt, l);
     fflush(stderr);
     va_end(l);
     Cilk_mutex_signal(&USE_MACHINE(dprintf_lock));
}

void Cilk_dprintf_machine(const char *fmt,...)
{
     va_list l;

     fprintf(stderr, "machine %2d: ", USE_MACHINE(Machine));
     va_start(l, fmt);
     vfprintf(stderr, fmt, l);
     fflush(stderr);
     va_end(l);
}

void Cilk_dprintf_global(const char *fmt,...)
{
     va_list l;

     /* ensure nice interleaving of debugging messages */
     wait_dprintf_lock();
     fprintf(stderr, "    global: ");
     va_start(l, fmt);
     vfprintf(stderr, fmt, l);
     fflush(stderr);
     va_end(l);
     Cilk_mutex_signal(&USE_MACHINE(dprintf_lock));
}

DECLARE_MACHINE_VARIABLE(static Cilk_mutex, die_lock, );

void Cilk_die(const char *fmt,...)
{
     va_list l;
     int success;

     success = Cilk_mutex_try(&USE_MACHINE(die_lock));

     wait_dprintf_lock();
     fprintf(stderr, "   (%2d,%2d): fatal error: ", USE_MACHINE(Machine), USE_PRIVATE(Process));
     va_start(l, fmt);
     vfprintf(stderr, fmt, l);
     va_end(l);

     {
       char c = 4;
       write(1, &c, 1);
       c = 5;
       write(2, &c, 1);
       while(1);        /* wait for kill from global process */
     }
     
     Cilk_mutex_signal(&USE_MACHINE(dprintf_lock));
}

void Cilk_die_machine(const char *fmt,...)
{
     va_list l;

     fprintf(stderr, "machine %2d: fatal error: ", USE_MACHINE(Machine));
     va_start(l, fmt);
     vfprintf(stderr, fmt, l);
     va_end(l);

     {
       char c = 4;
       write(1, &c, 1);
       c = 5;
       write(2, &c, 1);
       while(1);        /* wait for kill from global process */
     }
}

void Cilk_die_global(const char *fmt,...)
{
     va_list l;

     fprintf(stderr, "    global: fatal error: ");
     va_start(l, fmt);
     vfprintf(stderr, fmt, l);
     va_end(l);

     exit(1);
}

/***********************************************************\
 * Mapping of variable segments
\***********************************************************/

#if !CILK_MANAGE_SHARED_SEGMENT 
/*
 * if linker segments are not used, a portable version
 * of these two functions is provided in cilk.c
 */

/* 
 * map the global variables segment as shared
 */
void Cilk_map_shared_segment(void)
{
     extern _end_Cilk_shared;	/* linker_defined symbols */
     extern _begin_Cilk_shared;
     char name[1024];
     int fd;
     void *startmem, *endmem;

     strcpy(name, "/tmp/_Cilk_gvars_XXXXXX");
     mktemp(name);

     startmem = &_begin_Cilk_shared;
     endmem = &_end_Cilk_shared;

     /* write all shared data to the file */
     fd = open(name, O_RDWR | O_CREAT, 0600);
     CILK_CHECK(fd >= 0, ("Open error\n"));
     unlink(name);
     CILK_CHECK(write(fd, startmem, endmem - startmem) == endmem - startmem,
		("error in write\n"));

     /* map shared data to be shared */
     if (startmem != mmap(startmem, endmem - startmem, PROT_READ | PROT_WRITE,
			  MAP_SHARED | MAP_FIXED, fd, 0))
	  Cilk_die_global("couldn't map shared part to file.\n");
}
#endif

#if !CILK_MANAGE_PRIVATE_SEGMENT 
void Cilk_map_private_segment(void)
{
     /* variables are private by default */
}
#endif

void Cilk_map_machine_segment(void)
{
     extern _end_Cilk_machine;	/* linker_defined symbols */
     extern _begin_Cilk_machine;
     char name[1024];
     int fd;
     void *startmem, *endmem;

     strcpy(name, "/tmp/_Cilk_gvars_XXXXXX");
     mktemp(name);

     startmem = &_begin_Cilk_machine;
     endmem = &_end_Cilk_machine;

     /* write all machine data to the file */
     fd = open(name, O_RDWR | O_CREAT, 0600);
     if (fd < 0)
       Cilk_die_machine("Open error\n");
     unlink(name);
     if (write(fd, startmem, endmem - startmem) != endmem - startmem)
       Cilk_die_machine("error in write\n");

     /* map machine data to be shared */
     if (startmem != mmap(startmem, endmem - startmem, PROT_READ | PROT_WRITE,
			  MAP_SHARED | MAP_FIXED, fd, 0))
	  Cilk_die_machine("couldn't map machine part to file.\n");
}

/***********************************************************\
 * Malloc support
\***********************************************************/

/* Get a chunk of memory of size <len> that is mapped shared. */
void *Cilk_map_shared(size_t len)
{
     int fd;
     void *area;

     fd = open("/dev/zero", O_RDWR);

     CILK_CHECK(fd >= 0, ("Open error in Cilk_map_shared\n"));

     area = mmap(0, len, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
     if (area == MAP_FAILED)
          Cilk_die_global("mmap error in Cilk_map_shared -- "
			  "couldn't allocate %d bytes\n", len);

     close(fd);
     return area;
}

/***********************************************************\
 * Architecture-specific options
\***********************************************************/

/*
 * Parse any architecture-specific command-line options.  <argc>
 * and <argv> are the original arguments passed on the command line,
 * and <i> is the index of the first architecture-specific option (if any).
 * This function returns the index of the first user option, or <argc> if
 * there aren't any.
 */
int Cilk_arch_specific_options(int i, int argc, char *argv[])
{
     return i;
}

/***********************************************************\
 * Support for mutexen 
\***********************************************************/
#if CILK_USE_MEMORY_LOCKS

#if SPARC_V9
#define ACQUIRE(lock) Cilk_cas(lock, 0, 1)
#define RELEASE(lock) *lock = 0
#endif

#if SPARC_V8
#define ACQUIRE(lock) Cilk_swap(lock, 1)
#define RELEASE(lock) Cilk_swap(lock, 0)
#endif

/* initialize lock */
void Cilk_mutex_init(Cilk_mutex *lock)
{
     *lock = 1;
     RELEASE(lock);
}

/*
 *  Obtain lock <lock>.  Waits until lock is available.
 */
void Cilk_mutex_wait(Cilk_mutex *lock)
{
     if (ACQUIRE(lock) == 0)
          return;
     
     Cilk_enter_state(STATE_WAITING_FOR_LOCK);

     /* 
      * This hackery should reduce bus traffic, if the
      * cache is snooped correctly.
      */
     do {
	  while (*lock != 0);
     } while (ACQUIRE(lock) != 0);

     Cilk_exit_state(STATE_WAITING_FOR_LOCK);
}

/*
 *  Try to obtain lock <lock>.  Returns 1 if it succeeds and 0 if it fails.
 *  Does not wait.
 */
int Cilk_mutex_try(Cilk_mutex *lock)
{
     if (*lock != 0)		/*
				 * don't try for lock until we know we
				 * might succeed 
				 */
	  return 0;
     if (ACQUIRE(lock) == 0) {
	  return 1;
     } else
	  return 0;
}

/*
 *  Give up lock <lock>.
 */
void Cilk_mutex_signal(Cilk_mutex *lock)
{
     RELEASE(lock);
}

#else

/* initialize lock */
void Cilk_mutex_init(Cilk_mutex *lock)
{
     mutex_init((mutex_t *) lock, USYNC_PROCESS, NULL);
}

/*
 *  Obtain lock <lock>.  Waits until lock is available.
 */
void Cilk_mutex_wait(Cilk_mutex *lock)
{
     Cilk_enter_state(STATE_WAITING_FOR_LOCK);

     mutex_lock((mutex_t *) lock);

     Cilk_exit_state(STATE_WAITING_FOR_LOCK);
}

/*
 *  Try to obtain lock <lock>.  Returns 1 if it succeeds and 0 if it fails.
 *  Does not wait.
 */
int Cilk_mutex_try(Cilk_mutex *lock)
{
     if (mutex_trylock((mutex_t *) lock) == 0) {
	  return 1;
     } else {
	  return 0;
     }
}

/*
 *  Give up lock <lock>.
 */
void Cilk_mutex_signal(Cilk_mutex *lock)
{
     mutex_unlock((mutex_t *) lock);
}

#endif

/***********************************************************\
 * Ctrl-C handler
\***********************************************************/

static int nmachines;
static void kill_machines(int n);
static void ctrlc_handler(int sig)
{
  kill_machines(nmachines);
}

static void setup_ctrlc_handler(int nmach)
{
  nmachines = nmach;
  signal(SIGINT, ctrlc_handler);
}

/***********************************************************\
 * Cleanup signal handling
\***********************************************************/

static void cleanup_handler(int sig)
{
  CILK_ASSERT(sig == SIGUSR1);
  kill(0, SIGKILL);
}

static void setup_cleanup_handler(void)
{
  signal(SIGUSR1, cleanup_handler);
}

/***********************************************************\
 * Cache eviction signal handling
\***********************************************************/

DECLARE_MACHINE_VARIABLE(static pid_t *, children_pid, );

extern void Cilk_cache_evict_pages(void);

static void eviction_handler(int sig)
{
  CILK_ASSERT(sig == SIGUSR2);
  Cilk_cache_evict_pages();
}

static void setup_eviction_handler(void)
{
  sigset(SIGUSR2, eviction_handler);
}

void Cilk_send_eviction_notice(int i)
{
  kill(USE_MACHINE(children_pid)[i], SIGUSR2);
}

/***********************************************************\
 * Spawning worker processes
\***********************************************************/

#include <sys/wait.h>
#include <malloc.h>
#include <signal.h>
#include <sys/resource.h>

/*
 * Start forking processes: one process enters this function,
 * <Cilk_active_size> processes should exit this function immediately
 * and return 0.  Each returning process should set <Self> to a unique
 * number between 0 and <Cilk_active_size> - 1.  After the returning
 * processes complete (via Cilk_end_parallel()), the original process
 * should exit this function with a non-zero return value.
 */
void Cilk_create_children(const Options *options, void (*child)(const Options *, int))
{
     int i, n;
     pid_t pid;
     int stat;
     int error = 0;

     setup_cleanup_handler();

     USE_MACHINE(children_pid) = (pid_t *) Cilk_malloc_machine(options->permach * sizeof(pid_t));
     if (!USE_MACHINE(children_pid))
       Cilk_die_machine("couldn't malloc() children_pid\n");

     for (i = 0; i < options->permach; i++) {
	  if ((pid = fork()) < 0) {
	       Cilk_die_machine("can't fork");
	  } else if (pid == 0) {	/* child */
	       child(options, i);
	       exit(0);
	  } else {		/* parent */
	       USE_MACHINE(children_pid)[i] = pid;
	       /* Cilk_dprintf_machine("child %d %d\n", i, (int)pid); */
	  }
     }

     /* parent now waits for all children to complete */
     for (n = 0; n < options->permach; n++) {
	  pid = waitpid((pid_t) - 1, &stat, 0);

	  /* find completed child's # */
	  for (i = 0; i < options->permach; i++)
	       if (pid == USE_MACHINE(children_pid)[i])
		    break;
	  CILK_ASSERT(i < options->permach);

	  if (WIFEXITED(stat)) {
	       if (WEXITSTATUS(stat) == 0) {
		    continue;
	       } else
		    Cilk_dprintf_machine("process %2d - exited with non-zero value %d\n",
					 i, WEXITSTATUS(stat));
	  } else if (WIFSIGNALED(stat)) {
	       const char *core;
	       if (WCOREDUMP(stat)) {
		    core = " (core dumped)";
		    /* prevent others from overwriting core, if possible */
		    Cilk_mutex_try(&USE_MACHINE(die_lock));
	       } else
		    core = "";

	       /* 
	        * Print out understandable messages for common signals,
	        * and the signal number for the uncommon ones. 
	        */
	       switch (WTERMSIG(stat)) {
		   case SIGILL:
			Cilk_dprintf_machine("process %2d - Illegal instruction%s\n", i, core);
			break;
		   case SIGABRT:
			Cilk_dprintf_machine("process %2d - Abort or Assertion failed%s\n", 
				     i, core);
			break;
		   case SIGFPE:
			Cilk_dprintf_machine("process %2d - Floating exception%s\n", i, core);
			break;
		   case SIGKILL:
			Cilk_dprintf_machine("process %2d - Terminated%s\n", i, core);
			break;
		   case SIGBUS:
			Cilk_dprintf_machine("process %2d - Bus error%s\n", i, core);
			break;
		   case SIGSEGV:
			Cilk_dprintf_machine("process %2d - Segmentation fault%s\n", i, core);
			break;
		   default:
			Cilk_dprintf_machine(
			  "process %2d - signal %d (see /usr/include/sys/signal.h)"
					 "%s\n", i, WTERMSIG(stat), core);
			break;
	       }
	  }
	  /* 
	   * something went wrong with child i: abort the rest of the 
	   * children.  Sending a SIGKILL keeps them from dumping core.
	   */
	  for (i = 0; i < options->permach; i++)
	       kill(USE_MACHINE(children_pid)[i], SIGKILL);
	  error = 1;
     }
     
     if (error) {
          /*
	   * Kill self.  This will abort all of the
	   * other machines.
	   */
          kill(getpid(), SIGKILL);
     }
}

/***********************************************************\
 * Spawning worker machines
\***********************************************************/

void Cilk_start_machine(int argc, char *argv[])
{
  Options options;
  int machine;
  
  /*
   * Read options and machine ID from the master program.
   */
  read(0, &options, sizeof(options));
  read(0, &machine, sizeof(machine));
  
  options.argc = argc;
  options.argv = argv;
  
  Cilk_machine_main(&options, machine);
}



static int mconn[MAXMACHINES];  /* stdout of each slave */
static int merr[MAXMACHINES];   /* stderr of each slave */

static void kill_machines(int n)
{
  int i;
  char c = SIGUSR1;
  /* fprintf(stderr, "killing machines\n"); */
  for(i=0; i<n; i++)
    write(merr[i%n], &c, 1);
}


void Cilk_create_machines(Options *options)
{
  struct servent *se;
  int res;
  struct pollfd fds[2*MAXMACHINES];
  int i;
  char name[2*MAXPATHLEN+1];
  int n = options->nmach;
  char *executable = options->argv[0];
  char **names = Cilk_option_machine;
  int *sizes = Cilk_option_nproc;
  int done = 0;
  int sum, count;
  
  CILK_ASSERT(n <= MAXMACHINES);

  /*
   * Hack to get full pathname of executable
   */
  if (executable[0] == '/')
    {
      /* executable is an absolute path name - no need to mess with it */
      CILK_ASSERT(strlen(executable) <= MAXPATHLEN);
      strcpy(name, executable);
    }
  else
    {
      /* executable is a relative path name - prepend working directory */
      res = (int)getcwd(name, MAXPATHLEN+1);
      if (res == 0) { perror("can't get working directory"); exit(1); }
      CILK_ASSERT(strlen(name) + strlen(executable) + 1 <= MAXPATHLEN);
      strcat(name, "/");
      strcat(name, executable);
    }
  strcat(name, " -__machine__");
  for(i=1; i<options->argc; i++)
    {
      strcat(name, " ");
      strcat(name, options->argv[i]);
    }
  printf("executable: %s\n", name);fflush(stdout);
  
  /*
   * Start each slave, save a socket to it.
   */
  se = getservbyname("exec", "tcp");
  if (se == NULL)
    return;
  
  for(i=0; i<n; i++)
    {
      if (names[i] == NULL)
	{
	  names[i] = getenv("HOST");
	  if (names[i] == NULL)
	    names[i] = getenv("HOSTNAME");
	  if (names[i] == NULL)
	    { printf("can't get local host name!\n"); exit(1); }
	}
      res = rexec(&names[i], se->s_port, NULL, NULL, name, &merr[i]);
      if (res < 0) { printf("can't rexec to `%s'!\n", names[i]); exit(1); }
      mconn[i] = res;
      printf("started slave %d (on %s)\n", i, names[i]);fflush(stdout);
    }
  
  /*
   * Send options to each machine
   */
  sum = 0;
  for(i=0; i<n; i++)
    sum += sizes[i];
  count = 0;
  for(i=0; i<n; i++)
    {
      /*
       * Fill in per-machine options
       */
      options->start_proc_no = count;
      options->permach = sizes[i];
      /* options->heap_ptr filled in from shmem-distrib.c */
      options->nproc = sum;
      write(mconn[i], options, sizeof(*options));
      write(mconn[i], &i, sizeof(i));
      count += sizes[i];
    }
  
  /* initialize message-passing library */
  Cilk_message_passing_init_machine_globalproc(options, mconn);
  
  /*
   * For the rest of time, wait for data on any of the mconn[], and print it.
   */
  /* set reads from mconn[] to be nonblocking */
  for(i=0; i<n; i++)
    {
      int flags;
      flags = fcntl(mconn[i], F_GETFL, 0);
      CILK_ASSERT(flags >= 0);
      flags |= O_NONBLOCK;
      res = fcntl(mconn[i], F_SETFL, flags);
      CILK_ASSERT(res >= 0);

      flags = fcntl(merr[i], F_GETFL, 0);
      CILK_ASSERT(flags >= 0);
      flags |= O_NONBLOCK;
      res = fcntl(merr[i], F_SETFL, flags);
      CILK_ASSERT(res >= 0);
    }
  for(i=0; i<n; i++)
    {
      fds[i].fd = mconn[i];
      fds[i].events = POLLIN;
      fds[i+n].fd = merr[i];
      fds[i+n].events = POLLIN;
    }
  while(1)
    {
      res = poll(fds, 2*n, INFTIM);
      CILK_ASSERT(res > 0);
      
      for(i=0; i<2*n; i++)
	{
	  if (fds[i].revents & POLLIN)
	    while(1)
	      {
		char c;
		res = read(fds[i].fd, &c, 1);
		if (res == 0)
		  {
#if 1
		    printf("slave %d closed prematurely\n", i%n);
		    fds[i].fd = -1;
		    kill_machines(n);
		    return;
#endif
		    break;
		  }
		if (res == -1)
		  break;   /* all of the data for now */
		
		CILK_ASSERT(res == 1);
		if (c == 4)
		  {
		    /* printf("slave %d %s done\n", i%n, i<n ? "out":"err"); */
		    fds[i].fd = -1;
		    done++;
		    if (done == 2*n)
		      {
			kill_machines(n);
			return;
		      }
		    break;
		  }
		else if (c == 5)
		  {
		    /* bug occurred at machine */
		    fds[i].fd = -1;
		    done++;
		    kill_machines(n);
		    if (done == 2*n)
		      return;
		  }
		else if (i<n)
		  putc(c, stdout);
		else
		  putc(c, stderr);
		if (c=='\n') break;  /* ensure fairness */
	      }
	  if (fds[i].revents & ~POLLIN)
	    {
	      printf("something weird happened with slave %d %s\n", i%n, i<n ? "out":"err");
	      fds[i].fd = -1;
	    }
	}
    }
}

/***************************************************
 * Timer control
 ***************************************************/
static int openslashproc(int pid)
{
     char buf[64];
     int  fd;

     sprintf(buf, "/proc/%d", pid);
     if ( (fd = open(buf, 2)) < 0 )
	  perror("open /proc");

     return fd;
}

static void start_microaccounting(void)
{
     int fd, flags;

     fd = openslashproc(getpid());

     flags = PR_MSACCT;
     if (ioctl(fd, PIOCSET, &flags, sizeof (flags)) < 0) {
	  perror("ioctl set microstate acct");
     }
}


/**********************************************************
 * Priority control
 **********************************************************/
#define LOW_PRIORITY (-60)

DECLARE_PRIVATE_VARIABLE(static int, current_priority,);
void Cilk_lower_priority(void)
{
     struct pcparms params;
     struct tsparms *tsparams = (struct tsparms *) &params.pc_clparms;

     if (USE_PRIVATE(current_priority) != LOW_PRIORITY) {
	  params.pc_cid = PC_CLNULL;
	  priocntl(P_PID, getpid(), PC_GETPARMS, (caddr_t) & params);
	  USE_PRIVATE(current_priority) = tsparams->ts_upri = LOW_PRIORITY;
	  priocntl(P_PID, getpid(), PC_SETPARMS, (caddr_t) & params);
     }
}

void Cilk_raise_priority(void)
{
     struct pcparms params;
     struct tsparms *tsparams = (struct tsparms *) &params.pc_clparms;

     if (USE_PRIVATE(current_priority) == LOW_PRIORITY) {
	  params.pc_cid = PC_CLNULL;
	  priocntl(P_PID, getpid(), PC_GETPARMS, (caddr_t) & params);
	  USE_PRIVATE(current_priority) = 
	       tsparams->ts_upri = tsparams->ts_uprilim;
	  priocntl(P_PID, getpid(), PC_SETPARMS, (caddr_t) & params);
     }
}

/***********************************************************\
 * Fault handling
\***********************************************************/

#include <signal.h>
#include <sys/ucontext.h>

DECLARE_PRIVATE_VARIABLE(static struct sigaction, oldact, );
DECLARE_PRIVATE_VARIABLE(static struct sigaction, newact, );

extern int Cilk_fault_handler(void *addr, int write);
DECLARE_MACHINE_VARIABLE(extern int, pagesize, );

#define TRACE(x,y)

static void segv_handler(int sig, siginfo_t * info, ucontext_t * context)
{
     TRACE(TRACE_SIGNAL, ("got signal %d (%d)\n", sig, info->si_code));

     /* 
      * The si_core != SEGV_ACCERR is commented out, because
      * it only works with mprotect() and not with munmap().
      * See comments in Cilk_unmap_cache_page()
      */

     if (sig != SIGSEGV ||	/* info->si_code != SEGV_ACCERR || */
	 !Cilk_fault_handler(info->si_addr, 0)) {
	  /* 
	   * uninstall handler and return - user code will get real signal 
	   */
	  sigaction(SIGSEGV, &USE_PRIVATE(oldact), NULL);
	  Cilk_dprintf("segmentation fault at address %p\n", info->si_addr);
	  sleep(5);
	  return;
     }
     TRACE(TRACE_SIGNAL, ("return from signal handler\n"));
}

void Cilk_setup_fault_handler(void)
{
     sigaction(SIGSEGV, NULL, &USE_PRIVATE(oldact));	/* get signal stuff */
     USE_PRIVATE(newact) = USE_PRIVATE(oldact);		/* copy stuff       */
     USE_PRIVATE(newact).sa_sigaction = (void (*)()) segv_handler;
     USE_PRIVATE(newact).sa_flags |= SA_SIGINFO | SA_NODEFER;	/* modify stuff     */
     sigaction(SIGSEGV, &USE_PRIVATE(newact), NULL);	/* set signal stuff */
}

/***********************************************************\
 * Memory mapping
\***********************************************************/

/***********************************************************\
 * Alignment macros
\***********************************************************/

#define PAGE_ALIGNED_P(addr) ((addr) == ALIGN((addr), USE_MACHINE(pagesize)))

#include <sys/mman.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>

/*
 * Allocate a chunk of virtual memory of size <len>, and initialize it
 * to fault when referenced.  Returns pointer to the start of memory.
 */
void *Cilk_map_alloc(void *addr, size_t len)
{
     void *area;
     int fd = open("/dev/zero", O_RDWR);
     
     CILK_ASSERT(fd >= 0);
     area = mmap(addr, len, PROT_NONE, MAP_PRIVATE | MAP_NORESERVE | (addr == NULL ? 0:MAP_FIXED), fd, 0);
     if (area == MAP_FAILED)
	  Cilk_die("mmap error in Cilk_map_alloc\n");
     if (addr != NULL && area != addr)
          Cilk_die("couln't mmap same heap on each machine!\n");

     close(fd);
     return area;
}

void Cilk_unmap_region(void *addr, size_t size)
{
     if (size == 0) return;
     CILK_ASSERT(PAGE_ALIGNED_P(addr));
     /* Cilk_enter_state(STATE_USER0); */
#ifdef USE_MPROTECT
     /* 
      * I suspect mprotect() is broken on Solaris.
      * I think I found a case in which a page mapped PROT_NONE
      * is still accessible. Maybe the hardware is broken. -- athena
      */
     CILK_CHECK(mprotect(addr, size, PROT_NONE) == 0,
		("couldn't mprotect region! %p\n", addr));
#else
     /* 
      * we use munmap() for now. 
      */
     CILK_CHECK(munmap(addr, size) == 0,
		("couldn't munmap region! %p\n", addr));
     /* Cilk_exit_state(STATE_USER0); */
#endif
}


/* file descriptor of backing store file */
#define CACHE_NAME "_Cilk_cache"
DECLARE_MACHINE_VARIABLE(static int, cache_fd, );
DECLARE_MACHINE_VARIABLE(static void *, cache_base, );

/*
 * Note: all arguments of the functions below are guaranteed to be a
 * multiple of the system page size.
 */

/*
 * Allocate and set up cache area of <len> bytes.
 */
void *Cilk_setup_cache(size_t len, const Options *options)
{
     int junk = 42;
     char name[PATH_MAX + 100];
     
     sprintf(name, "%s/%s_%d_%d", options->cachedir, CACHE_NAME, (int)getpid(), USE_MACHINE(Machine));

     /* create the file */
     USE_MACHINE(cache_fd) = open(name, O_RDWR | O_CREAT, 0600);

     if (USE_MACHINE(cache_fd) < 0)
       Cilk_die_machine("Open error for %s in Cilk_setup_cache\n", name);

     /* delete temporary cache file */
     unlink(name);

     /* extend cache file to be <len> bytes long */
     /* Is there a better way to do this? */

     if (lseek(USE_MACHINE(cache_fd), len - sizeof(int), SEEK_SET) != len - sizeof(int))
       Cilk_die_machine("error in lseek!\n");

     if (write(USE_MACHINE(cache_fd), &junk, sizeof(int)) != sizeof(int))
       Cilk_die_machine("error in write!\n");
     
     USE_MACHINE(cache_base) = mmap(0, len, PROT_READ | PROT_WRITE, MAP_SHARED, USE_MACHINE(cache_fd), 0);
     if (USE_MACHINE(cache_base) == MAP_FAILED)
       Cilk_die_machine("couldn't map cache into memory!\n");
     
     return USE_MACHINE(cache_base);
}

/*
 * Map a page of memory of memory [<addr>, <addr>+<pagesize>-1] to 
 * the page <cache_addr> of the cache page <cache_addr>.  Accesses
 * are restricted to those allowed by <mode>.
 */
void Cilk_map_cache_page(void *addr, void *cache_addr, int mode)
{
     CILK_ASSERT(PAGE_ALIGNED_P(addr));
     CILK_ASSERT(PAGE_ALIGNED_P(cache_addr));
     
     if (!(mode & CILK_MAP_NOSTAT)) Cilk_enter_state(STATE_USER1);
     switch(mode & 0xF)
       {
       case CILK_MAP_FAULTING:
	 CILK_CHECK(munmap(addr, USE_MACHINE(pagesize)) == 0,
		    ("couldn't munmap page! %p\n", addr));
	 break;
       case CILK_MAP_READONLY:
	 if (addr != mmap(addr, USE_MACHINE(pagesize), PROT_READ,
			  MAP_SHARED | MAP_FIXED, USE_MACHINE(cache_fd),
			  cache_addr - USE_MACHINE(cache_base)))
	   Cilk_die("couldn't map read-only.\n");
	 break;
       case CILK_MAP_READWRITE:
	 if (addr != mmap(addr, USE_MACHINE(pagesize), PROT_READ | PROT_WRITE,
			  MAP_SHARED | MAP_FIXED, USE_MACHINE(cache_fd),
			  cache_addr - USE_MACHINE(cache_base)))
	  Cilk_die("couldn't map to read-write.\n");
	 break;
       }
     if (!(mode & CILK_MAP_NOSTAT)) Cilk_exit_state(STATE_USER1);
     
     if (!(mode & CILK_MAP_NOSTAT)) Cilk_enter_state(STATE_USER2);
     /*
      * Force OS to install new mapping in page table.  We do this
      * here only because we want to know how much time this takes,
      * and we want to attribute the time spent in the system to
      * this mapping itself, not to the user's code.
      */
     if (mode == CILK_MAP_READWRITE)
       {
	 /* 
	  * CAS is treated as both a read and a write by the memory
	  * system, so it forces the page to be mapped writeable.
	  * However, this call to CAS doesn't actually change memory,
	  * so it is safe.
	  */
#if SPARC_V9
	 Cilk_cas(addr, 0, 0);
#else
	 /* no operation has the right semantics!  Oh well.  Just
	  * make sure it is mapped readable at least.
	  */
	 *((volatile int *)addr);
#endif
       }
     else if (mode == CILK_MAP_READONLY)
       {
	 *((volatile int *)addr);
       }
     if (!(mode & CILK_MAP_NOSTAT)) Cilk_exit_state(STATE_USER2);
}

/**********************************************************
 * Initialization of the module
 **********************************************************/

void Cilk_arch_specific_init_global(Options *options)
{
     options->pagesize = sysconf(_SC_PAGESIZE);

     start_microaccounting();
     
     setup_ctrlc_handler(options->nmach);
}

void Cilk_arch_specific_init_machine(const Options *options)
{
     /* make sure all machines use the same page size */
     CILK_ASSERT(options->pagesize == sysconf(_SC_PAGESIZE));

     USE_MACHINE(pagesize) = options->pagesize;

     Cilk_mutex_init(&USE_MACHINE(dprintf_lock));
     Cilk_mutex_init(&USE_MACHINE(die_lock));

     start_microaccounting();
}

void Cilk_arch_specific_init_local(const Options *options)
{
     USE_PRIVATE(current_priority) = -1;
     start_microaccounting();
     setup_eviction_handler();
}

void Cilk_arch_specific_done_machine(const Options *options)
{
}

void Cilk_arch_specific_done_global(const Options *options)
{
}
