
/*
 * $Id: barrier.c,v 1.2.2.2 1997/11/10 22:27:53 randall Exp $
 */

/*
 * $Log: barrier.c,v $
 * Revision 1.2.2.2  1997/11/10 22:27:53  randall
 * Updated some interfaces, moved some more socket initialization
 * into message-udp.c
 *
 * Revision 1.2.2.1  1997/06/13 22:33:05  randall
 * Added distributed version to the repository under
 * branch "distributed"
 *
 * Revision 1.2  1997/03/01 17:42:37  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

/*
 * This file contains the code for barriers
 */

#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-internal.h>

/***********************************************************\
 * Code for barriers
\***********************************************************/
DECLARE_MACHINE_VARIABLE(static Cilk_mutex, barrier_lock, );
DECLARE_MACHINE_VARIABLE(static int, global_count, );  /* only used on machine 0 */
DECLARE_MACHINE_VARIABLE(static int, local_count, );
DECLARE_MACHINE_VARIABLE(volatile int, barrier_release, );
DECLARE_PRIVATE_VARIABLE(int, barrier_direction, );

typedef struct
{
  int direction;
} BarrierRequest;

typedef struct
{
  int direction;
} BarrierReply;

static void barrier_reply(BarrierReply *reply)
{
  USE_MACHINE(barrier_release) = reply->direction;
}

static void barrier_request(BarrierRequest *req)
{
  int i;
  
  CILK_ASSERT(USE_MACHINE(Machine) == 0);
  
  Cilk_mutex_wait(&USE_MACHINE(barrier_lock));
  USE_MACHINE(global_count)++;
  if (USE_MACHINE(global_count) == USE_MACHINE(Machines))
    {
      USE_MACHINE(global_count) = 0;
      Cilk_mutex_signal(&USE_MACHINE(barrier_lock));
      /* Cilk_dprintf("passed barrier\n"); */
      for(i=0; i<USE_MACHINE(Machines); i++)
	{
	  BarrierReply *reply;
	  reply = Cilk_alloc_msg_buf(sizeof(BarrierReply), i);
	  reply->direction = req->direction;
	  Cilk_rpc(i, barrier_reply, reply, sizeof(*reply));
	}
    }
  else
    Cilk_mutex_signal(&USE_MACHINE(barrier_lock));
}

#include <stdio.h>
void Cilk_barrier(void)
{
  CILK_ASSERT(USE_PRIVATE(barrier_direction) == USE_MACHINE(barrier_release));
  
  USE_PRIVATE(barrier_direction) = !USE_PRIVATE(barrier_direction);
  
  Cilk_mutex_wait(&USE_MACHINE(barrier_lock));
  
  USE_MACHINE(local_count)++;
  if (USE_MACHINE(local_count) == USE_MACHINE(Processes))
    {
      BarrierRequest *req;
      USE_MACHINE(local_count) = 0;
      Cilk_mutex_signal(&USE_MACHINE(barrier_lock));
      req = Cilk_alloc_msg_buf(sizeof(BarrierRequest), 0);
      req->direction = USE_PRIVATE(barrier_direction);
      Cilk_rpc(0, barrier_request, req, sizeof(*req));
    }
  else
    Cilk_mutex_signal(&USE_MACHINE(barrier_lock));
  
  while(USE_MACHINE(barrier_release) != USE_PRIVATE(barrier_direction))
    { Cilk_poll(); Cilk_yield(); }
}

void Cilk_barrier_init_machine(const Options *options)
{
     /* initialize the barrier code */
     Cilk_mutex_init(&USE_MACHINE(barrier_lock));

     USE_MACHINE(global_count) = 0;
     USE_MACHINE(local_count) = 0;
     USE_MACHINE(barrier_release) = 0;
}

void Cilk_barrier_init_local(const Options *options)
{
  USE_PRIVATE(barrier_direction) = 0;
}
